/** \file k644p.c  Routines for setting up and using ATMEGA644P onchip peripherals
 *
 *  This file is for a simple demo.  Optionally, a 3-axis accelerometer can be
 *  connected to PORT A pins 0 through 5.   Pins 0, 1 and 2 are the analog output
 *  of channels x, y, and z.   Pins 3 and 4 are the gain select, and pin 5 is the
 *  accelerometer sleep function.
 *
 *  A hardware board with the Freescale MMA7260Q chip on it is availble from
 *  sparkfun.com for around $20.  It is a breeze to attach this to the Crumb644
 *  from chip45.com.
 *
 *  Check http://www.BoldInventions.com for more info on this.
 *
 */
#include "kcmd/kcmdIncludes.h"
#include "k644p.h"


/**
 * g_tmr0 is a structure containing variables related to timer0
 */
TIMERSTATE_t g_tmr0;


/**
 * g_tmr1_nTick is the tick count for Timer1 
 */
uint16_t g_tmr1_nTick;

/**
 * init_timer0
 * 
 *  Initializes timer0 hardware
 */
void initTimer0(void)
{
    g_tmr0.tickCount=0;  
    g_tmr0.secondCount=0;
    g_tmr0.delayCounter=0;

	/* Set up CTC mode - Clear Timer on Compare. */
	/* Leave COM0Ax bits zero - we leave OC0A disconnected. */
	TCCR0A = _BV(WGM01);


	/* Set the prescaler to divide by 256.   8000000/256 = 31250 Hz */
	TCCR0B = _BV(CS02);


	/* Now we divide (8000000/256) by TMR0_TICKS_PER_SEC to find out what  */
	/* value to put in the OCR0A.  The timer will interrupt the CPU */
	/* and clear itself whenever the count reaches OCR0A.          */
	/*   So for 1008 ticks per second, we have (8000000/256)/1008 - 1 = 30 */
	/*   This will mean an interrupt every 1.6 milliseconds. */
	OCR0A = F_CPU/TMR0_TICKS_PER_SEC/256 - 1;

	/* Clear Output Compare Match A flag, so interrupt will get called when */
	/* the timer reaches the OCR0A value. */
	TIFR0 = _BV(OCF0A);

	/* Unmask the interrupt, so an interrupt will get called when */
	/* the timer reaches the OCR0A value. */
	TIMSK0 = _BV (OCIE0A);	

}


ISR (TIMER0_COMPA_vect)
{
	// Work on delay if we are waiting on a delay of some kind
        if(g_tmr0.delayCounter) g_tmr0.delayCounter--;

	/* increment rtc tick and if overflowed, increment seconds */
	g_tmr0.tickCount++;
	if (g_tmr0.tickCount == TMR0_TICKS_PER_SEC) {
		g_tmr0.secondCount++;
		g_tmr0.tickCount = 0;
	}
}

/**
 * set_tmr0_delay() sets the delay count which is in the global
 * g_tmr0 structure.  The Timer0 ISR will decrement this count
 * once every timer0 interrupt, until it reaches zero.
 * Code wishing to delay by a certain number of milliseconds can
 * set this delay and then keep calling get_tmr0_delay() until
 * it is zero again.
 */
void set_tmr0_delay(uint16_t value)
{
	cli();
	g_tmr0.delayCounter=value;
	sei();
}


/**
 * get_tmr0_delay() gets the current value of the delay count
 * from the global g_tmr0 structure.   The tmr0 ISR will be
 * decrementing delay at the tick rate, until the delay reaches
 * zero.
 */
uint16_t get_tmr0_delay(void)
{
	uint16_t uiRet;

	cli();  // Turn off interrupts
        uiRet=g_tmr0.delayCounter;
	sei();
	return uiRet;
}


/**
 * get_tmr0_ticks() gets the current value of the tmr0 tick counter.
 */
uint16_t get_tmr0_ticks(void)
{
	uint16_t uiRet;

	cli();  // Turn off interrupts
        uiRet=g_tmr0.tickCount;
	sei();
	return uiRet;
}


/**
 * wait_msec() waits for a nMsec to elapse before returning.
 * It requires interrupts to be enabled and tmr0 to be running.
 * Otherwise it will hang forever.
 */
void wait_msec(uint16_t nMsec)
{
	cli();
	   g_tmr0.delayCounter=nMsec;
	sei();

	// Now we could go into a lower power mode
	// while we wait.
	
	// But for now we just go into a tight spinning loop.
	do
	{
	} while( 0!=get_tmr0_delay());
}




/**
 * init_timer1
 * 
 *  Initializes timer1 hardware to run at ADC_SAMPLE_RATE, and generate interrupts.
 */
void initTimer1(void)
{
    g_tmr1_nTick=0;

	PRR &= ~_BV(PRTIM1);  // Power up Timer1

	// Also setup timer1 in CTC mode because we want high accuracy on the 240 samp/sec rate
	// on the a/d convertor
	TCCR1A =  0;  // WGM1 and WGM0 are zero, COM1A0,1 = 0 COM1B0,1 = 0
	TCCR1B =   0x08; // WGM2 = 1, Clock Scale = 0 to stop timer for now.
	OCR1A  = F_CPU/ADC_SAMPLE_RATE;
	TIMSK1 = _BV(OCIE1A);  // Enable interrupts when count matches timer 1;
	TCCR1B = 0x09; // Turn prescaler to 1 to start the counter counting.
}


ISR (TIMER1_COMPA_vect)
{
	/* increment rtc tick and if overflowed, increment seconds */
	g_tmr1_nTick++;

	// Start converting, we will get an ADC interrupt when the conversion is complete
	 ADCSRA=0xCE;  // Enable A/d (0x80) | ADSC (ADC STart Conversion) | ADC Int Enable (0x8) | prescaler=64 (0x6)
}




/**
 * g_adc is the global data structure for holding ADC data and state. 
 */
ADC_DATA g_adc;

/**
 * getAdcDataPtr() is a helper function which simply retrieves the
 * current state of the ADC data structure. 
 */
ADC_DATA *getAdcDataPtr(void)
{
   return &g_adc;
}

/**
 * initADC(void)
 *
 * Initializes the ADC for converting on PORTA pins 0, 1 and 2.
 **/
void initADC(void)
{
    int i;

    /* Initialize the global data structure. */
    g_adc.m_index=0;  // counts up with each sample, modulo N_ADC_FILTERPT
    g_adc.m_chan=0;   // goes from 0 to N_ADC_CHAN
	for(i=0; i<N_ADC_ARRAYSZ; i++)
	{
      g_adc.m_rawdata[i]=0;
      g_adc.m_avgdata[i]=0;
    }
    g_adc.m_nConv=0; // Keep track of the number of conversions for fun

	PRR &= ~_BV(PRADC);  // Power up ADC

	
	/*
	 * config PORTA bits 0-2 for ADC operation
	 */

	DDRA &= ~(_BV(DDA0) | _BV(DDA1) | _BV(DDA2));  // Bits 0,1 and 2 to inputs
    
    /* Turn off pullup resistors for PORT A bits 0-2 */
	PORTA= ~(_BV(PORTA0) | _BV(PORTA1) | _BV(PORTA2)); 


    /* When the analog level on a digital input pin is near the logic threshold, */
	/* the digital input circuitry can waste lots of power.  We must disable the */
	/* digital input buffer for the pins we are going to use for ADC input.      */
	DIDR0 |= (_BV(ADC0D) | _BV(ADC1D) | _BV(ADC2D));  

    /* Dividing the 8Mhz clock by 64 will give us an ADC clock of 125 MHz        */
	/* The doc says we should run the ADC clock at 50kHz to 200kHz               */
    ADCSRA=0x6;  // Disable and set clock divider to divide by 64

    /* The top two bits define the voltage reference to use.  We will use VCC    */
	/* to get the full range of conversion.  The doc states we need to put a     */
	/* external capacitor from AREF to ground to reduce noise.                   */
	/*                                                                           */
	/* We are going to set the ADLAR bit, which will left justify the data bits  */
	/* which is good because we are going to divide the signal down during       */
	/* filtering.                                                                */
    ADMUX = 0x60;  // VCC as reference, left justfied, portA0 single ended source

    /* We will run in free run mode, with the Timer1 ISR manually starting the   */
	/* A/D conversions.                                                          */

    /* Free Run Mode, ADATE (Auto Trig En will be off so this is moot anyway. )  */
     ADCSRB= 0x0; 

	 ADCSRA=0x8E;  // Enable A/d (0x80) | ADC Int Enable (0x8) | prescaler=64 (0x6)
}




/**
 * the ADC Interrupt Service Routine.
 */
ISR(ADC_vect)
{
   uint16_t adcval;
   int16_t iTmp;
   uint8_t index;
   uint8_t iChan;
   uint8_t i;

   index=g_adc.m_index;
   iChan=g_adc.m_chan;

   i= index + (iChan << ADC_FILTERSHIFT);

   adcval = ADC;  // Get 16 bit value

   g_adc.m_rawdata[i] = adcval; // Store most recent raw data

   iTmp = (int16_t) g_adc.m_avgdata[i];  // Retrieve avg data


   /* If N_ADC_POW_SHIFTDIV == 6, then we will divide the difference between */
   /* the avg value and the new value by 64, and then add it to the average. */
   /* That way, it takes about 64 data collection events for the average to  */
   /* catch up to the current ADC value.  This is our digital filter to      */
   /* reduce sample to sample noise.                                         */
   g_adc.m_avgdata[i] += ((int16_t) adcval - iTmp)>>N_ADC_POW_SHIFTDIV;


   /* Go to next channel */
   iChan++;

   /* If that was the last channel, then we should go back to waiting for    */
   /* another timer interrupt.                                               */
   if(N_ADC_CHAN <= iChan)
   {
      g_adc.m_chan=0;  /* Go back to the first channel */
	  ADMUX = 0x60;  // VCC as reference, left justfied, portA0 single ended source
      g_adc.m_index = (index+1) & ADC_FILTERMASK;  // Advance index
	  g_adc.m_nConv++;  // Count conversions as a diagnostic

      // And now are done, we just wait for the Timer1 Interrupt to start
	  // the next conversion...

   } else
   {
   	  ADMUX = 0x60 | (iChan);  // Set the ADC Mux
      g_adc.m_chan=iChan;    // Update the channel number


  	 // Start converting, we will get another ADC interrupt when the conversion is complete
	 ADCSRA=0xCE;  // Enable A/d (0x80) | ADSC (ADC STart Conversion) | ADC Int Enable (0x8) | prescaler=64 (0x6)

   }
}


/**
 * initGpio() initalizes the GPIO lines.
 */
void initGpio(void)
{
    /*
	 * This configuration of PortA is for an interface to the MMA7260Q 
	 * triple-axis accelerometer board from sparkfun.com
	 *
	 * I have connected PortA 0, 1, 2 to the analog data out lines
	 * for x, y, and x axes.  Gain Sel1 and Gain Sel2 are connected to
	 * PortA3, and Port A4.   The sleep signal is connected to  PortA5.
	 * So PortA 3,4, 5 should all be outputs.
	 */

	/*
	 * config PORTA bits 3-5 for Digital Output
	 */
	 	PINA |= _BV(PINA3) | _BV(PINA4) | _BV(PINA5);

	DDRA |= (_BV(DDA3) | _BV(DDA4) | _BV(DDA5));  // Bits 3,4 and 5 to outputs
    
    /* Put to all zeros except Pin 5 which is the sleep pin.  We want */
	/* sleep to stay high.  The Gain select we'll start out at zero   */
	/* for minimum gain. */
	PORTA &= ~(_BV(PORTA3) | _BV(PORTA4)); 
	PORTA |= _BV(PORTA5);


}






