/** \file main.c
 *  \brief A simple program to test out the portable kcmd command processor.
 *
 *  This is a simple example program test out the kcmd command shell.  It is
 *  written for the AVR_ATmega644P microprocessor.   It will use UART0 for
 *  the default input and output.
 * 
 *  This demo can make use of a 3-axis accelerometer connected to PORT A,
 *  pins 0 thru 5.  However, it runs fine if nothing is connected to PORTA.
 *
 *  It does require that the serial port UART0 is connected to a PC or some
 *  other system so the user can type commands and see the displayed output.
 */

#include "kcmd/kcmdIncludes.h"
#include "k644p/k644p.h"
#include "my_commands.h" /* WARNING- only include this line in your main .c file. */
#include "main.h"


/* Make sure F_CPU is defined or else code won't work right. */
#ifndef F_CPU
#error "F_CPU not defined"
#endif

/* ---- Private Constants and Types --------------------------------------- */

#define BAUDRATE 19200 /**< Fixed UART baud rate at 19200 */


/**
 * Macro to calculate UBBR from XTAL and baudrate
 */
#define UART_BAUD_SELECT(baudRate,xtalCpu) (((float)(xtalCpu))/(((float)(baudRate))*16.0)-1.0+0.5)


/**
 * init_uart() initializes the usart.
 *
 */
static void
init_uart(void)
{
	/* Setup usart for required baud rate and transmitting data only */
	REG_UART0_BAUD_LOW = UART_BAUD_SELECT(BAUDRATE,F_CPU);
	REG_UART0_CONTROL   = _BV(BIT_UART0_TX_EN) | _BV(BIT_UART0_RX_EN);
}


/**
 * uartPutChar - sends a character out on the UART
 *
 * Simple spinning loop uart send function which waits forever for the uart to 
 * become ready before sending one char. 
 *
 * @param c the character to send.
 */
void uartPutChar(char c)
{  
	/* wait until previous character is transmitted before sending the next one */
	loop_until_bit_is_set(REG_UART0_STATUS, BIT_UART0_TX_RDY);
	REG_UART0_DATA = c;                       
}

/**
 * uartGetChar - waits for a character to be received and returns it.
 *
 *  This function waits forever for a received character to appear in the uart's
 *  receive buffer.
 *
 *  Note: If it is desired to have 'idle loop' code which executes whenever
 *  the system is waiting for the user to hit a key, this routine is the 
 *  place for it.
 *
 *  return the character received
 */
char uartGetChar(void)
{
  char c;
  loop_until_bit_is_set(REG_UART0_STATUS, BIT_UART0_RX_COMPLETE);
  c = REG_UART0_DATA;
  return c;
}


/**
 * uartPutString() writes a string to the uart.
 *
 * @param pStr is a pointer to the beginning of a null-terminated string.
 */
void uartPutString(char *pStr)
{
  char c;
  char * addr;

  addr=(char *) pStr;
  while ((c = *(addr++)))
    uartPutChar(c);
}

/**
 * uartPutString() writes a flash-stored string to the uart.
 *
 * @param pStr is a pointer to the beginning of a null-terminated string.
 */
void uartPutString_P(const char *pStr)
{
  char c;
  unsigned short addr;

  addr=(unsigned short) pStr;
  while ((c = __LPM(addr++)))
    uartPutChar(c);
}

/**
 * uartIsCharAvail
 *
 * return 0 if no characters available, 1 if there is a char avail.
 */
int uartIsCharAvail(void)
{
  int iRet;
  iRet= (REG_UART0_STATUS & _BV(BIT_UART0_RX_COMPLETE) ? 1 : 0);
  return iRet;
}


/**
 * main 
 * 
 * Initializes the uart and goes into an infinite loop waiting for characters.
 *
 * return It is silly to return a value, but this the format for main.
 */
int main(void)
{
	int iExitFlag;


	/* Set the LED GPIO line to be an output. */
	REG_LEDA_DDR  |= _BV(BIT_LEDA);

	/* initializes I/O for this application */
	init_uart();

    /* init and start timer0 */
    initTimer0();

    /* initialize the ADC */
    initADC();

    /* Initialize the timer1 and start converting. */
    initTimer1();

    /* Set up the GPIO lines. */
	initGpio();


    /* Initialize the kcmd command processor */
       init_kcmd( 
	     PSTR("\r\nPROMPT> "),
		 PSTR("\r\n\nKCMD Command Processor Demo\r\n"
		 "\n  from www.boldinventions.com\r\n"
		 "\r\n  Type help for help\r\n"),
	    uartGetChar, uartIsCharAvail, uartPutChar, uartPutString  );


     /*
	  * Loop forever getting input lines from the input device and 
      * processing the command lines.
	  */
     do
     {
        iExitFlag=process_kcmds(1);
     } // Do-while loop
     while(1!=iExitFlag);

	return 0;
}
