/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file kcmd_basic_commands.h contains the source code for the basic
 * set of commands.
 */
#ifndef KCMD_BASIC_COMMANDS_H__
#define KCMD_BASIC_COMMANDS_H__


/****************************************************************************
 * Yes, we are putting compiled code in a header file!
 * This is because we need control over the order of compilation for the 
 * linked list macros to work.
 ****************************************************************************/

/****************************************************************************
 *  This file has the basic commands defined.  Each command has its own
 *  C function.   The functions all take one parameter, and that is a pointer
 *  to a KCMD structure.   The functions all return nothing.
 *   
 *  Using the KCMD stucture, the command functions can do input and output
 *  to the user, as well as taking arguments from the stack or leaving output
 *  on the stack for other commands to use as input.
 *
 *  void kcmd_Easter(KCMD *pKCMD)
 *  {
 *    print_kccstr(pKCMD, KCCSTR("\r\negg.\r\n"));
 *  }
 *  
 *  This command just prints the word 'egg' to the user, with a CR 
 *    before and after.   The 'print_kccstr()' function is specific for
 *    printing constant (KCCSTR) strings ONLY.   
 *
 *  For printing a string stored in memory, you can call the helper functions
 *    which in the printf style.  Note that the 'format' parameter of the
 *    kcmd_Printf_x() functions are always type KCCSTR, meaning that they
 *    must be stored in the Program or Flash memory if the platform uses
 *    that feature.
 *
 *  The following example uses a formatted print to show the user the values
 *    of two global variables.
 *
 *  void kcmd_memstat(KCMD *pKCMD)
 *  {
 *     extern int g_nBytesAllocated;
 *     extern int g_nMaxBytesAllocated;
 *     kcmd_Printf_2(pKCMD, 
 *         KCCSTR("\r\n Cur Bytes Allocated: %d  Max: %d\r\n"),
 *         g_nBytesAllocated, 
 *         g_nMaxBytesAllocated);
 *  }
 ****************************************************************************/

/* If we are in a c++ environment, make sure the compiler knows these are */
/* straight C functions. */
#ifdef __cplusplus
extern "C" {
#endif

/****************************************************************************
 * The basic system commands are defined here.
 ****************************************************************************/

void kcmd_Easter(KCMD *pKCMD)
{
    print_kccstr(pKCMD, KCCSTR("\r\negg.\r\n"));
}

MAKE_KCMD_COMMAND( easter, kcmd_Easter, "[] -- [] Celebrates Easter", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(easter)





void kcmd_StartApp(KCMD *pKCMD)
{
    print_kccstr(pKCMD, KCCSTR("\r\nWhat App?.\r\n"));
}
MAKE_KCMD_COMMAND( startApp, kcmd_StartApp, "[] -- [] do not use", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(startApp)


KCCHAR KCCATTR g_str_stackEmpty[]=KT("\r\nStack Empty!\r\n");
KCCHAR KCCATTR g_str_crlf[] = KT("\r\n");
KCCHAR KCCATTR g_kcmd_msg_stack_underflow[] = KT("\r\nStack Underflow!\r\n");

void kcmd_DotStack(KCMD *pKCMD)
{
   if(pKCMD->iStack==0)
   {
    print_kccstr(pKCMD, g_str_stackEmpty);
   } else
   {
      int i;
    print_kccstr(pKCMD, g_str_crlf);
      for(i=0; i<pKCMD->iStack; i++)
      {
         if(pKCMD->base==10)
         {
           kcmd_Printf_1(pKCMD, KCCSTR("%d "), pKCMD->stack[i]);
         } else
         {
           kcmd_Printf_1(pKCMD, KCCSTR("%x "), pKCMD->stack[i]);
         }
      }
    print_kccstr(pKCMD, g_str_crlf);
   }
}
MAKE_KCMD_COMMAND_5( dotStack, ".stack",  kcmd_DotStack, "[] -- [] types out entire stack", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(dotStack)


void kcmd_Dot(KCMD *pKCMD)
{
   if(pKCMD->iStack==0)
   {
    print_kccstr(pKCMD, g_str_stackEmpty);
   } else
   {
      int i;
      i=kcmd_stackPop(pKCMD);
         if(pKCMD->base==10)
         {
           kcmd_Printf_1(pKCMD, KCCSTR("%d "), i);
         } else
         {
           kcmd_Printf_1(pKCMD, KCCSTR("%x "), i);
         }
   }
}
MAKE_KCMD_COMMAND_5( dot, ".", kcmd_Dot, "[n] -- [] types top of stack", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(dot)


void kcmd_Drop(KCMD *pKCMD)
{
   kcmd_stackPop(pKCMD);
}
MAKE_KCMD_COMMAND( drop, kcmd_Drop, "[n] -- [] drops top of stack", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(drop)


void kcmd_At(KCMD *pKCMD)
{
   int value;
   value=kcmd_stackAt(pKCMD);
   kcmd_stackPush(pKCMD, value );
}
MAKE_KCMD_COMMAND_5( at, "@",  kcmd_At, "[n] -- [memAt(n)] dereferences ptr", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(at)


void kcmd_Store(KCMD *pKCMD)
{
   KINTPTR_t iAddr;
   int iValue;
   if( pKCMD->iStack<2) 
   { 
      print_kccstr(pKCMD, g_kcmd_msg_stack_underflow);
   } else
   {
     iAddr=kcmd_stackPop(pKCMD);
     iValue= (int) kcmd_stackPop(pKCMD);
     *((volatile int *) iAddr)=iValue;
   }
}
MAKE_KCMD_COMMAND_5( store, "!", kcmd_Store, "[n] [addr] -- [] Writes n to address", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(store)


void kcmd_Question(KCMD *pKCMD)
{
   kcmd_stackQuestion(pKCMD);
}
MAKE_KCMD_COMMAND_5( question, "?", kcmd_Question, "[n] -- [] derefences pts and types value", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(question)



void kcmd_Dup(KCMD *pKCMD)
{
   kcmd_stackDup(pKCMD);
}
MAKE_KCMD_COMMAND( dup, kcmd_Dup, "[n] -- [n] [n] Duplicates top of stack", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(dup)

void kcmd_Swap(KCMD *pKCMD)
{
   kcmd_stackSwap(pKCMD);
}
MAKE_KCMD_COMMAND( swap, kcmd_Swap, "[x][y] -- [y][x] swaps top two elems on stk", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(swap)

void kcmd_Over(KCMD *pKCMD)
{
   kcmd_stackOver(pKCMD);
}
MAKE_KCMD_COMMAND( over, kcmd_Over, "[a][b] -- [a][b][a] dups 2nd stk elem", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(over)


void kcmd_Plus(KCMD *pKCMD)
{
   int i1, i2;
   i1=kcmd_stackPop(pKCMD);
   i2=kcmd_stackPop(pKCMD);
   kcmd_stackPush(pKCMD, i1+i2 );
}
MAKE_KCMD_COMMAND_5( plus, "+",  kcmd_Plus, "[a][b] -- [a+b] Adds top two stk elems", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(plus)


void kcmd_Minus(KCMD *pKCMD)
{
   int i1, i2;
   i1=kcmd_stackPop(pKCMD);
   i2=kcmd_stackPop(pKCMD);
   kcmd_stackPush(pKCMD, i2-i1 );
}
MAKE_KCMD_COMMAND_5( minus, "-",  kcmd_Minus, "[a][b] -- [a-b] subtracts two top stk elems", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(minus)


void kcmd_Times(KCMD *pKCMD)
{
   int i1, i2;
   i1=kcmd_stackPop(pKCMD);
   i2=kcmd_stackPop(pKCMD);
   kcmd_stackPush(pKCMD, i2*i1 );
}
MAKE_KCMD_COMMAND_5( times, "*", kcmd_Times, "[a][b] -- [a*b] multiplies top 2 elems", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(times)


void kcmd_Divide(KCMD *pKCMD)
{

   int i1, i2;
   i1=kcmd_stackPop(pKCMD);
   i2=kcmd_stackPop(pKCMD);
   if(i1>0)
   {
      kcmd_stackPush(pKCMD, i2/i1 );
   } else 
   {
          kcmd_Printf_0(pKCMD, KCCSTR("\r\nError: Divide by zero!\r\n"));
   }
}
MAKE_KCMD_COMMAND_5( divide, "/", kcmd_Divide, "[a][b] -- [a/b] divides", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(divide)


void kcmd_memstat(KCMD *pKCMD)
{
extern int g_nBytesAllocated;
extern int g_nMaxBytesAllocated;
   kcmd_Printf_2(pKCMD, KCCSTR("\r\n Cur Bytes Allocated: %d  Max: %d\r\n"),
      g_nBytesAllocated, g_nMaxBytesAllocated);
}
MAKE_KCMD_COMMAND( memstat, kcmd_memstat, "[] -- [] types out memory status info", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(memstat)


void kcmd_NumBase(KCMD *pKCMD)
{
   int i1;
   i1=kcmd_stackPop(pKCMD);
   switch(i1)
   {
      case 10:
          pKCMD->base=10;
          kcmd_Printf_0(pKCMD, KCCSTR("\r\nNumber Base set to 10.\r\n"));
      break;

      case 16:
          pKCMD->base=16;
          kcmd_Printf_0(pKCMD, KCCSTR("\r\nNumber Base set to 16.\r\n"));
      break;

      default:
          kcmd_Printf_0(pKCMD, KCCSTR("\r\nNumber Base can only be set to 10 or 16.\r\n"));
      break;
   }
}
MAKE_KCMD_COMMAND( base, kcmd_NumBase, "[n] -- [] set number base to n", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(base)



char kcmd_TCHAR_or_dot(char ch)
{
	if( (ch < ' ') || (ch > 0x7f) ) ch='.';
	return ch;
}


/**
 * Convert an integer to a hex string.   Returns a pointer to the next 
 * position to write a character.
 */
char *kcmd_hex_to_string(char *cbuf, int iHappy, int nChar, KINTPTR_t value)
{
  char *pRet;
  int i;
  int iNum;
  char ch;
  int szInt;

  szInt=sizeof(int);
  pRet=cbuf;
  for(i=0; i<(szInt*2); i++)
  {
     iNum= (int) ((value >> ((szInt*8)-4)) & 0xf);
     ch=iNum + '0';
     if(iNum>9) ch = iNum-10+'A';
     if( (2*szInt)-i-nChar-1 < 0 ) *pRet++ = ch;
     value=value << 4;  // Move next nibble into the high order 4 bits
  }
  *pRet=0;  // Terminate string
  return pRet;
}



void kcmd_Dump(KCMD *pKCMD)
{
   if(pKCMD->iStack<2)
   {
          kcmd_Printf_0(pKCMD, KCCSTR("\r\nDUMP needs [addr] [nBytes] on stack!\r\n"));
   } else
   {
      unsigned int i;
      KINTPTR_t iaddr;
      int itmp;
      unsigned int nBytes;
      KINTPTR_t uiAddr;
      KINTPTR_t adrListStart;
      unsigned int nBytesFirstLine;
      unsigned int nBytesLastLine;
      unsigned int nMiddleLines;
      KCHAR *pByte;
      char *pouch;
      USER_IO *pUserIO;
      char cbuf[16];


      pUserIO=pKCMD->pUserIO;

      itmp=kcmd_stackPop(pKCMD);
      uiAddr=kcmd_stackPop(pKCMD);

      if( itmp<1 )
      {
          kcmd_Printf_1(pKCMD, KCCSTR("\r\nToo Few bytes in count: %d"), itmp);
          return;
      }
      if( itmp>16000 )
      {
          kcmd_Printf_1(pKCMD, KCCSTR("\r\nToo Many bytes in count: %d"), itmp);
          return;
      }

      nBytes= (unsigned int) itmp;
      kcmd_crlf(pKCMD);

	  for(i=0; i<(3+sizeof(KINTPTR_t)*2); i++) pUserIO->pFuncPutc(pKCMD->conio_context, ' ', 300);

      kcmd_Printf_0(pKCMD, KCCSTR("|00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f\r\n"));

	  for(i=0; i<(3+sizeof(KINTPTR_t)*2); i++) pUserIO->pFuncPutc(pKCMD->conio_context, '-', 300);

	  kcmd_Printf_0(pKCMD, KCCSTR("-------------------------------------------------|----------------\r\n"));



      adrListStart=uiAddr & ~0xf;
      nBytesFirstLine= (unsigned int) adrListStart+0x10 - uiAddr;
      if(nBytesFirstLine > nBytes) nBytesFirstLine=nBytes;
      itmp=nBytes+uiAddr-adrListStart-0x10;
      if(itmp<0)
      {
         nMiddleLines=0;
         nBytesLastLine=0;
      } else
      {
         nMiddleLines=itmp>>4;
         nBytesLastLine=itmp&0xf;
      }

      // Draw first line:
      pouch=kcmd_hex_to_string(cbuf, 1, (sizeof(KINTPTR_t)*2), adrListStart);
	  *pouch=0; // terminate string
	  print_kccstr(pKCMD,  KCCSTR("0x" ));
      pUserIO->pFuncPuts(pKCMD->conio_context, cbuf);
	  print_kccstr(pKCMD,  KCCSTR(" |"));
	  for(i=0; (int) i<uiAddr - adrListStart; i++)
	  {
	      print_kccstr(pKCMD,  KCCSTR("   ") );
	  }
      for(iaddr=uiAddr; iaddr<(KINTPTR_t) (uiAddr+nBytesFirstLine); iaddr++)
      {
         pByte=(KCHAR *) iaddr;
         kcmd_hex_to_string(cbuf, 1, 2, (int) (*pByte & 0xff));
         cbuf[2]=' ';
         cbuf[3]=0;
         pUserIO->pFuncPuts(pKCMD->conio_context, cbuf);
      }
	  itmp=(int)adrListStart +0x10 -(int) uiAddr-(int)nBytesFirstLine;
	  if( itmp>0)   // Skip over to end if on last line
	  {
			   for(i=0; i<(unsigned int) itmp; i++)
			   {
		          pUserIO->pFuncPuts(pKCMD->conio_context, KT("   "));
			   }
	  }
	  for(i=0; (int) i<uiAddr - adrListStart; i++)
	  {
		  pUserIO->pFuncPutc(pKCMD->conio_context, ' ', 1000);
	  }

      for(iaddr=uiAddr; iaddr<(KINTPTR_t) (uiAddr+nBytesFirstLine); iaddr++)
      {
         pByte=(KCHAR *) iaddr;
         pUserIO->pFuncPutc(pKCMD->conio_context, kcmd_TCHAR_or_dot(*pByte),1000);
      }
      kcmd_crlf(pKCMD);

      // Draw Middle Lines
      if(nBytesFirstLine<nBytes)
      {
         unsigned int iLine;
         unsigned int nBytesThisLine;
         for(iLine=0; iLine<nMiddleLines+1; iLine++)
         {
           pouch=kcmd_hex_to_string(cbuf, 1, (sizeof(KINTPTR_t)*2), adrListStart + ((iLine+1)<<4));
		   *pouch=0;  // terminate string
	        print_kccstr(pKCMD,  KCCSTR("0x" ));
           pUserIO->pFuncPuts(pKCMD->conio_context, cbuf);
	        print_kccstr(pKCMD,  KCCSTR(" |"));
           nBytesThisLine=16;
           if(iLine==nMiddleLines) nBytesThisLine=nBytesLastLine;
           for(i=0; i<nBytesThisLine; i++)
           {
              pByte=(KCHAR *) (adrListStart+((iLine+1)<<4)+i);
              kcmd_hex_to_string(cbuf, 1, 2, (int) (*pByte & 0xff));
              cbuf[2]=' ';
              cbuf[3]=0;
              pUserIO->pFuncPuts(pKCMD->conio_context, cbuf);
           }
	       if(iLine==nMiddleLines)   // Skip over to end if on last line
		   {
			   for(i=0; i<16-nBytesLastLine; i++)
			   {
		          pUserIO->pFuncPuts(pKCMD->conio_context, KT("   "));
			   }
		   }
           for(i=0; i<nBytesThisLine; i++)
           {
              pByte=(KCHAR *) (adrListStart+((iLine+1)<<4)+i);
              pUserIO->pFuncPutc(pKCMD->conio_context, kcmd_TCHAR_or_dot(*pByte), 1000);
           }
             kcmd_crlf(pKCMD);;
         }
      }
      kcmd_crlf(pKCMD);

   }
}
MAKE_KCMD_COMMAND( bdump, kcmd_Dump, "[addr][count] -- [] displays memory contents", g_str_system );
#undef PREVNAME
#define PREVNAME G_ME(bdump)


#ifdef __cplusplus
} /* extern "C" ... */
#endif



#endif /* KCMD_BASIC_COMMANDS_H__ */
