/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file kcmd.c is main code for the kcmd simple command processor.
 *
 */

#include "kcmdIncludes.h"

const KCMD_CMD_LIST_ELEM KCSTRUCTATTR g_kcmd_help;

/****************************************************************************
 *  Some structures local only to this C file are defined below.  They are  *
 * meant to be elements of a list.                                          *
 ****************************************************************************/

/**
 * CATNAMES is a container for creating a list of category names.
 */
typedef struct struct_CATNAMES
{
	int nCategories;
	KCCHAR *name_list[KCMD_MAX_CATEGORIES];
} CATNAMES;

/****************************************************************************
 * Some global strings stored in Flash (on systems which support it) for    *
 * convenience and efficency.  system and easter are the basic categories   *
 * help and the help help text are here too, to make the original static    *
 * structures functional even before the add_basic_commands() runs.         *
 ****************************************************************************/

KCCHAR KCCATTR g_str_system[] =  KT("system");


KCCHAR KCCATTR g_str_bye[] = KT("bye");
KCCHAR KCCATTR g_helptext_bye[] = KT("[] -- [] Exits cmd processor");

void kcmd_Bye(KCMD *pKCMD)
{
    print_kccstr(pKCMD, KCCSTR("\r\nBYE!\r\n"));
          pKCMD->flags |= KCMD_FLAG_EXIT;
}




/* Define the first command to show up in the linked list. */
const KCMD_CMD_LIST_ELEM KCSTRUCTATTR g_kcmd_bye = { NULL, g_str_bye, kcmd_Bye, g_helptext_bye, g_str_system }; 


/**
 * kcmd_getCmdFromName traverses list of cmd looking for one that matches
 * the given string.  If
 * none of the existing cmds match, then returns a null.
 */
KCMD_CMD_LIST_ELEM* kcmd_getCmdFromName(KCHAR *pStrCmdName)
{
   KCMD_CMD_LIST_ELEM *pCmd=(KCMD_CMD_LIST_ELEM *) &g_kcmd_help;

   do
   {  
      if( KCC_strlen(pCmd->pcmdName) == strlen(pStrCmdName))
	  {
	    if(!KCC_strncmp(pStrCmdName, pCmd->pcmdName,  KCC_strlen(pCmd->pcmdName)))
        {
           break;
        }
	  }
      pCmd=(KCMD_CMD_LIST_ELEM *) pCmd->pNext;
   } while (NULL != pCmd);
   return pCmd;
}




/**
 * is_this_cat_already_in_list is a helper function to help make
 * the list of unique category names.  Returns 1 if the pCatName
 * is already in the container list, or 0 if it isn't.
 */
int is_this_cat_already_in_list( CATNAMES *pCatNames, KCCHAR *pCatName)
{
	int i;
	int iRet=0;
	for(i=0; i<pCatNames->nCategories; i++)
	{
		if( pCatNames->name_list[i] == pCatName )
		{
			iRet=1;
			break;
		}
	}
	return iRet;
}

/**
 * Run through the list of all commands and create a list of pointers
 * to unique category names.  For example if there are 12 commands in
 * the 'system' category, and 4 commands in the 'GPIO' category, then
 * this function would put a pointer to the string 'system' and a
 * pointer to the string 'GPIO' in the CATNAMES structure.
 */
void kcmd_make_category_list( CATNAMES *pCatNames )
{

   KCMD_CMD_LIST_ELEM *pCmd=(KCMD_CMD_LIST_ELEM *) &g_kcmd_help;
   pCatNames->nCategories=0;
   do
   {  
      if( !is_this_cat_already_in_list(pCatNames, pCmd->pCategory) )
	  {
		  pCatNames->name_list[pCatNames->nCategories]=pCmd->pCategory;
		  pCatNames->nCategories++;
		  if(KCMD_MAX_CATEGORIES <= pCatNames->nCategories) break;
	  }
      pCmd=(KCMD_CMD_LIST_ELEM *) pCmd->pNext;
   } while (NULL != pCmd);
}


/**
 * kcmd_Help is a kcmd command which picks up the 
 * next word on the line and interprets it as a category name.
 * if no matching category is found, it types the list of 
 * categories for the user.   If one is found, it lists all 
 * the commands in that category.
 */
void kcmd_Help(KCMD *pKCMD)
{
   USER_IO *pUserIO=pKCMD->pUserIO; 
   KCHAR *pCmdLine=pKCMD->pCmdLine;
   int nameLen;
   KCHAR *spaces;
   CATNAMES catNames;
   KCCHAR *pCat;
   int i;
   unsigned int uiLen;
   KCMD_CMD_LIST_ELEM *pCmd=NULL;


   // const char * to the name of the category the user wants to list
   pCat=NULL;

   // TODO: must get rid of this malloc and find a way to do this
   // without wasting RAM.
   spaces=KCMD_MALLOC(sizeof(KCHAR)*25);

   // Make a list of categories to list the help on.
   kcmd_make_category_list( &catNames );


   /* Find out which category we are interested in. */
   if(pCmdLine!=NULL)
   {
     KCHAR *token;
     
     token=NULL;

     /* Call once to get the number of characters in token. */
     uiLen=0;
     if(NULL != parse_text_line(pCmdLine, NULL, &uiLen))
     {

         if(uiLen>0)
         {
             token=KCMD_MALLOC(sizeof(KCHAR) * uiLen);
         }
     }

     if(NULL != token)
     {
//     KCHAR *rest;
       parse_text_line(pCmdLine, token, &uiLen);
       pCmdLine[0]=0;  /* Leave flag that we are using the command line. */
       if(strlen(token)) 
	   {
        // If we got here, then the user has a category name after the
	    // word 'help' on the command line.
       
        // We need to look through all the category names and find
	    // the matching one.
	    pCat=NULL;
        for(i=0; i<catNames.nCategories; i++)
	    {
		   if(!KCC_strncmp(catNames.name_list[i], token, KCC_strlen(catNames.name_list[i])))
		   {
			   pCat=catNames.name_list[i];
			   break;
		   }
	    }
        if(NULL==pCat)
        {
		   print_kccstr(pKCMD, KCCSTR("\r\n Help category not found: "));
           pUserIO->pFuncPuts(pKCMD->conio_context, token);
		   kcmd_crlf(pKCMD);
        }
	   }
     }
	 if(NULL != token) KCMD_FREE(token);
   } else
   {
		 // If they did not type a category name on the command line after the word 'help'
		 // we end up here.
		 // If there is only one category, then we just want to list that category.  If there
		 // is more than one, then we want to tell them to type a category name after the world 'help'
		 if( 1 == catNames.nCategories )
		 {
            // Default value for the category is the first one on the list
            pCat=catNames.name_list[0];
		 }
   }

   /* If no category found, list the categories. */
   if(NULL==pCat)
   {
     int nInLine=0;
     int i;
     print_kccstr(pKCMD, KCCSTR("\r\n For help, type the word help and then one of the following keywords: \n\r"));
     
     for(i=0; i<catNames.nCategories; i++)
     {
	   pCat= catNames.name_list[i];
	   print_kccstr(pKCMD,  pCat);
       pUserIO->pFuncPuts(pKCMD->conio_context, KT(" "));
       nInLine += (int) KCC_strlen(pCat);
       if(nInLine>=(nLINEBUF-1))     // If we reach the end of the line, generate a CRLF to keep the line from getting too long.
       {
         kcmd_crlf(pKCMD);
         nInLine=0;
       }
     } 
     kcmd_crlf(pKCMD);
	 KCMD_FREE(spaces);
     return;
   }

   // If we arrived here, then the user must have typed 'help' followed by a valid category name which
   // now means that pCat points to a valid category.  So now we just type out all the 
   // commands which match this category.


   pCmd=(KCMD_CMD_LIST_ELEM *) &g_kcmd_help;
   do
   {
     int nSpaces;
     int j;
     if( pCmd->pCategory == pCat)
     {
       nameLen=(int) KCC_strlen(pCmd->pcmdName);
       kcmd_crlf(pKCMD);
       print_kccstr(pKCMD, pCmd->pcmdName);
       for(j=0; j<24; j++) spaces[j]=' ';
       nSpaces=24-nameLen;
       if(nSpaces < 0) nSpaces=0;
       spaces[nSpaces]=0;

       pUserIO->pFuncPuts(pKCMD->conio_context, spaces);

       if(pUserIO->pFuncKbhit(pKCMD->conio_context )) break;  /* Break if user hits a key. */
       print_kccstr(pKCMD, pCmd->pHelpText);
     }
     pCmd= (KCMD_CMD_LIST_ELEM *) pCmd->pNext;
   } while(NULL!=pCmd);

   kcmd_crlf(pKCMD);
   while(pUserIO->pFuncKbhit(pKCMD->conio_context )) pUserIO->pFuncGetc(pKCMD->conio_context, NULL, 10); /* Eat any keys that are in buffer. */
   KCMD_FREE(spaces);
}




/**
 * kcmd_cmdWelcome prints a welcome string when the command processor starts up. 
 */
void kcmd_cmdWelcome(KCMD *pKCMD)
{
   // If the user did not provide a welcome string when she called kcmd_cmdInit()
   // send out a default one.
   if(NULL==pKCMD->pWelcomeString)
   {
	KCCHAR *pLine0= KCCSTR("\r\nKCMD Command Processor");
	KCCHAR *pLine1= KCCSTR("\r\n For cmd list type help \r\n");

    print_kccstr(pKCMD, pLine0);
	print_kccstr(pKCMD, pLine1);
	} else
	{
	   print_kccstr(pKCMD, pKCMD->pWelcomeString);
	}
}




/**
 * Initializes the KCMD line processor.  This involves allocating memory for the 
 * command line processor structures and initializing them, and making the default list
 * of commands.
 *   The caller passes in a pointer to a USER_IO structure which gives the command line
 * processor the ability to request input and send output.  The prompt String allows the 
 * caller to set the prompt to something other than the default 'KCMD>' prompt.  
 *  The inputTimeout is passed to the input and output functions, and could be ignored
 * depending on the implementation of the USER_IO routines.
 *   This function returns a void pointer which is actually a pointer to the KCMD stucture,
 * The caller should hold onto this pointer, and then call kcmd_cmdExec(void *kcmd_context)
 * repeatedly while examining the returned integer to see if the user wanted to quit, or
 * to give error messages as appropriate.
 */
void *kcmd_cmdInit(
        USER_IO *pUserIO, 
        void *conio_context, 
	    KCCHAR *promptString,
	    KCCHAR *pWelcomeString, 
	    int inputTimeout)
	  {
      int i;
      KCMD *pKCMD;
     
      pKCMD=KCMD_MALLOC(sizeof(KCMD));
      pKCMD->flags=0;
      pKCMD->base=10;
      pKCMD->pUserIO=pUserIO;
      pKCMD->conio_context=conio_context;
      pKCMD->inputTimeoutInMS=inputTimeout;
	  pKCMD->pWelcomeString=pWelcomeString;
      for(i=0; i<N_CMDVM_STACK; i++) 
      {
         pKCMD->stack[i]=0;
      }
      pKCMD->iStack=0;
	  pKCMD->pPromptString=promptString;
      kcmd_cmdWelcome(pKCMD); // Print welcome message
      return (void *)  pKCMD;

}

/**
 * Takes a command name and the rest of the command line and searches the command list
 * for the command, and executes it.  Returns 0 if command was a number or a good command.
 * Can also return an exit flag, or a flag indicating we are to ignore the rest of the 
 * the command line, or an error flag.
 */
int kcmd_cmdParseWord(KCMD *pKCMD, KCHAR *token, KCHAR *rest)
{
   int iRet=KCMD_RFLAG_BADCOMMAND;
   int iErr;
   int i;
   KCMD_CMD_LIST_ELEM* pCmd;


   if(pKCMD==NULL) return -1;
   if(pKCMD->pUserIO==NULL) return -1;

   /* If the token is a number, put it onto the stack. */
  i=str_to_num(token, &iErr);
  if(iErr>-1)
  {
     kcmd_stackPush(pKCMD, i);
    iRet=0;
  } else
  {
     /* If not, parse it to see if it is an executable command. */

    pCmd = kcmd_getCmdFromName(token);
    if(NULL != pCmd)
    {
       iRet=0;
       pKCMD->pCmdLine=rest;
       pCmd->pFunkcmd_(pKCMD);
       if(pKCMD->flags & KCMD_FLAG_EXIT) return KCMD_RFLAG_USEREXIT;
       if(NULL==rest) return KCMD_RFLAG_IGNOREREST;
       if(rest[0]==0) return KCMD_RFLAG_IGNOREREST;  /* Check if cmd used the rest of the line. */
       if(pKCMD->flags & KCMD_RFLAG_STACKOVERFLOW) return KCMD_RFLAG_STACKOVERFLOW;
       if(pKCMD->flags & KCMD_RFLAG_STACKUNDERFLOW) return KCMD_RFLAG_STACKUNDERFLOW;
    } // if(NULL != pCmd)
  } // If(not a number)

  return iRet;
}

/**
 * Calls the console IO input function to see if there is a line of input available yet.
 * If there is not, returns with KCMD_RFLAG_NOINPUT.
 * If there is, the command is run with possible lines written out to console output.  The return
 * value can indicate error conditions like bad-command, or stack underflow/overflow, or
 * that the user has requested an exit of the command shell.
 *   For ConsoleIO functions which don't actually process the IO but simply queue up data, the 
 *   caller will need to call the appropriate functions to move data before or after each call
 *   to this cmd line interpreter. 
 **/
int kcmd_cmdExec(void *cmd_context, int bPrintPrompt)
{
   KCHAR *token;
   KCHAR *rest;
   int iRet;
   KCHAR srcbuf[nLINEBUF];
   int len;
   unsigned int uiLen;
   
   KCMD *pKCMD=(KCMD *) cmd_context;
   USER_IO *pUserIO=pKCMD->pUserIO; 
   void *conio_context=pKCMD->conio_context;

   pKCMD=(KCMD *) cmd_context;

   token=NULL;  // We make sure it is null, because we will try to free it later
                // And if it uninitialize, we will call free on it = very bad
   // If there is no input available, let caller know.
   iRet=KCMD_RFLAG_NOINPUT;
   if(bPrintPrompt)
   {
	print_kccstr(pKCMD, pKCMD->pPromptString);
   }
   len=pUserIO->pFuncGets(conio_context, srcbuf, nLINEBUF, pKCMD->inputTimeoutInMS); /* Get characters. */
   if(0<len)
   {
     pKCMD= pKCMD;


     token=NULL;
     iRet=0;
     rest=srcbuf;
     while( rest!=NULL ) 
     {
       iRet=KCMD_RFLAG_NOINPUT;
       token=NULL;
       uiLen=0;
       if(NULL!=parse_text_line(rest, NULL, &uiLen))
       {
           if(uiLen>0)
           {
               token=KCMD_MALLOC( uiLen );
           }
       }
       if(NULL != token)
       {
         rest=parse_text_line(rest, token, &uiLen);
         if(strlen(token)) {
//          print_kccstr(pKCMD, KCCSTR("\r\nCallnig kcmd_parseWord : "));
//		  pKCMD->pUserIO->pFuncPuts(pKCMD->conio_context, token);
          iRet= kcmd_cmdParseWord(pKCMD, token, rest);
//kcmd_Printf_1(pKCMD, KCCSTR("\r\nkmcd_cmdParseWord rtned: %d\r\n"), iRet);
          if(iRet==KCMD_RFLAG_USEREXIT) break;
          if(iRet== KCMD_RFLAG_BADCOMMAND) break;
          if(iRet== KCMD_RFLAG_STACKOVERFLOW) break;
          if(iRet== KCMD_RFLAG_STACKUNDERFLOW) break;
          if(iRet== KCMD_RFLAG_IGNOREREST) break;
         } 
         KCMD_FREE(token);
         token=NULL;
       }
     }

	 /* Clear any error conditions and clear the stack if there was an error. */
	 switch(iRet)
	 {
	 case KCMD_RFLAG_BADCOMMAND:
	 case KCMD_RFLAG_STACKOVERFLOW:
	 case KCMD_RFLAG_STACKUNDERFLOW:
		 pKCMD->flags=0;
		 pKCMD->iStack=0;
		 break;
	 default:
		 break;
	 } // switch(iRet);


   } // while(NULL != rest)
   if(NULL!=token) KCMD_FREE(token);

   return iRet;
}





