/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file kcmdSimpleInterface.c  - A simple command processor
  *
  */
#include "kcmdIncludes.h"


/**
  * KCMD_SIMPLE_IO_FUNC_PTRS is a container for function pointers 
  * for simple character IO.   See kcmdSimpleInterface.h for more
  * info about what the functions do.
  */
typedef struct struct_simple_io
{
    KCMDSIMPLE_GETCH m_pFuncGetch;
	KCMDSIMPLE_KBHIT m_pFuncKbHit;
	KCMDSIMPLE_PUTCH m_pFuncPutch;
	KCMDSIMPLE_PUTSTR m_pFuncPutStr;
} KCMD_SIMPLE_IO_FUNC_PTRS;


/**
  * g_simple_io_functions is a global container for the simple
  * function pointers.   The call to init_kcmd will initialize
  * them with function pointers provided by outside code. For
  * example, if it is desired to use a serial port for communication,
  * the main application would configure the serial port, and 
  * have simple IO functions that perform these tasks.  Alternatively,
  * the input functions might be reading a keyboard, while the output
  * functions draw on a LCD display.
  */
KCMD_SIMPLE_IO_FUNC_PTRS g_simple_io_functions;


/**
  * The USER_IO function is another container of IO routines with more
  * functionality.  These have the ability to timeout instead of waiting
  * forever for input, or space for output.  They have some error
  * handling ability.  This is the structure required for the command
  * processor.  However, this simple interface will make fake functions
  * out of the simpler IO to get the job done.  The call to init_kcmd()
  * will fill this global structure in.
  */
USER_IO g_uio;

/**
 * g_pkcmd is a pointer to a KCMD structure which holds the state of 
 * a command processor.  This means the stack and the input buffer of
 * text for parsing, flags and the current number base, etc.
 * In the more general case, there could be several instances of the
 * command processor, but for the simplest we just keep the kcmd context
 * in one global variable.
 *  This global variable get initialized and set up by init_kcmd();
 */
void *g_pkcmd;



/*
 * Global variables for keeping track of how much RAM we are using up. *
 */
int g_nBytesAllocated=0;
int g_nMaxBytesAllocated=0;

void simpleCHOUT(void *pContext, KCHAR ch)
{
    g_simple_io_functions.m_pFuncPutch(ch);
}

/**
  * kcmd_malloc() keeps track of how much memory has been allocated.
  *  to do this it must stored the size of the allocation in the memory
  * block itself, so that kcmd_free(ptr) can determine the size of the 
  * block which is getting freed.
  *  To help debug, if the malloc fails, it prints a message to the user. 
  */
void *kcmd_malloc( unsigned int nBytes )
{
    void *pRet;
	int *pInt;
	KCMD *pKCMD;

	pKCMD= (KCMD *) g_pkcmd;

	pInt = (int *) malloc( nBytes + sizeof(int) );
	if(NULL == pInt)
	{
	   kcmd_Printf_0( pKCMD, KCCSTR("\r\nPANIC: Malloc Failure. Tried to alloc  "));
	   int_to_string(simpleCHOUT, NULL, 0, 0,  nBytes);
       pRet=0;
	} else
	{
	   *pInt = nBytes;
	   pRet= (void *) (pInt+1);
	   g_nBytesAllocated+=nBytes;
	   if(g_nBytesAllocated > g_nMaxBytesAllocated) g_nMaxBytesAllocated=g_nBytesAllocated;
	}
	return pRet;
}

/**
  * kcmd_free() does a sanity check and prints a message if we get a bad
  * ptr, or if our length argument seems corrupted.
  */
void kcmd_free( void *ptr )
{
   int *pInt;
	KCMD *pKCMD;

	pKCMD= (KCMD *) g_pkcmd;


   if(NULL==ptr)
   {
     kcmd_Printf_0( pKCMD, KCCSTR("\r\nPANIC: attempt to free a NULL ptr\r\n"));
   }

   pInt = ((int *) ptr) - 1;
   if( 0 >= *pInt )
   {
     kcmd_Printf_0( pKCMD, KCCSTR("\r\nPANIC free() got a block of 0 or negative sz.\r\n"));
   }
   g_nBytesAllocated -= *pInt;
   free((void *) pInt);
}



KCHAR simpUio_getc(void *context, int *pInt, int iTimeout)
{
    if(NULL != pInt ) *pInt=1;  // Leave flag that no err happened.
    return (KCHAR) g_simple_io_functions.m_pFuncGetch();
}

int simpUio_putc(void *context, KCHAR ch, int iTimeout)
{
    g_simple_io_functions.m_pFuncPutch(ch);
    return 1;
}

int simpUio_kbhit(void *context)
{
    return g_simple_io_functions.m_pFuncKbHit();
}

int simpUio_outrdy(void *context)
{
    return 1;
}

int simpUio_gets(void *context, KCHAR *strBuf, int szBuf, int iTimeout)
{
    int bDone=false;
    KCHAR ch;
    int index=0;
	int iErr;
    do
    {
      ch=simpUio_getc(context, &iErr, iTimeout);

      switch(ch)
      {
        case '\n':  // Simply ignore linefeeds
          break;

        case '\r':  // Carriage return finishes the line.
            simpUio_putc(context, ' ', iTimeout);
            bDone=true;
          break;

        case '\b':  // Backspace moves the current recording spot back one.
          if(index>0)
		  {
			  simpUio_putc(context, ch, iTimeout);
              simpUio_putc(context, ' ', iTimeout);
              simpUio_putc(context, ch, iTimeout);
              strBuf[index]=0;
              index--;
		  }
          break;
        
        default:
          simpUio_putc(context, ch, iTimeout);
          if(index < szBuf)
          {
            strBuf[index]=ch;
            index++;
          }
          break;
      }
    } while( !bDone);
    strBuf[index]=0; // Add terminating null.
    return index;
}

void simpUio_puts(void *context, KCHAR *ch)
{
    g_simple_io_functions.m_pFuncPutStr(ch);
}

int simpUio_stop(void *context)
{
    return 1;
}

int simpUio_start(void *context)
{
    return 1;
}

void * simpUio_setup(void *context)
{
    return NULL;
}

int simpUio_release(void *context)
{
    return 1;
}


void init_uio( USER_IO *pUIO)
{
    pUIO->pFuncGetc = simpUio_getc;
    pUIO->pFuncGets = simpUio_gets;
    pUIO->pFuncKbhit = simpUio_kbhit;
    pUIO->pFuncOutRdy = simpUio_outrdy;
    pUIO->pFuncPutc = simpUio_putc;
    pUIO->pFuncPuts = simpUio_puts;
    pUIO->pFuncRelease = simpUio_release;
    pUIO->pFuncSetup = simpUio_setup;
    pUIO->pFuncStart = simpUio_start;
    pUIO->pFuncStop= simpUio_stop;
}


/**
  * init_kcmd() initializes the IO structure for our input and output
  * and initializes a context for the command processor, and stores it in
  * a global variable so the caller doesn't have to hold onto it.
  */


void init_kcmd(KCCHAR *pPrompt, 
               KCCHAR *pWelcomeString, 
               KCMDSIMPLE_GETCH pFuncGetch, 
			   KCMDSIMPLE_KBHIT pFuncKbHit,
			   KCMDSIMPLE_PUTCH pFuncPutch,
			   KCMDSIMPLE_PUTSTR pFuncPutStr)
{
   g_simple_io_functions.m_pFuncGetch = pFuncGetch;
   g_simple_io_functions.m_pFuncKbHit = pFuncKbHit;
   g_simple_io_functions.m_pFuncPutch = pFuncPutch;
   g_simple_io_functions.m_pFuncPutStr = pFuncPutStr;
   init_uio( &g_uio);
   g_pkcmd = kcmd_cmdInit(&g_uio, NULL, pPrompt, pWelcomeString, 100);
//   kcmd_Printf_1( (KCMD *) g_pkcmd, KCCSTR("kcmd_initted, pKCMD: 0x%0x\r\n"), (int) g_pkcmd);
}


/**
  * process_kcmds() gets a line of input from the user and processes the commands.
  * returns an integer.  0=done ok, 1=done ok but user wants exit. other positive
  * integers means there was a recoverable error.   To use, call this in a 
  * loop, and exit the loop when this function returns 1.  All other values
  * indicate it should be called again.   Call with iFlagPrompt=1 to print 
  * a command prompt.
  */
int process_kcmds(int iFlagPrompt)
{
      int iRet;
	  int iExitFlag=0;
	  KCMD *pKCMD;

	  pKCMD=(KCMD *) g_pkcmd;

      iRet=kcmd_cmdExec(g_pkcmd, iFlagPrompt!=0 );
	  
//kcmd_Printf_1(pKCMD, KCCSTR("\r\nkmcd_cmdExec rtned: %d\r\n"), iRet);

      if(!iExitFlag)
      {
        switch(iRet)
        {
            case KCMD_RFLAG_USEREXIT:
		    iExitFlag=1;
		    break;
            case KCMD_RFLAG_BADCOMMAND:
			   kcmd_Printf_0( (KCMD *) g_pkcmd, KCCSTR("\r\n Error Bad Command\r\n"));
	       break;
            case KCMD_RFLAG_STACKOVERFLOW:
				kcmd_Printf_0( (KCMD *) g_pkcmd, KCCSTR("\r\n Error Stack OVerflow\r\n"));
	       break;
            case KCMD_RFLAG_STACKUNDERFLOW:
			   kcmd_Printf_0( (KCMD *) g_pkcmd, KCCSTR("\r\n Error Stack Underflow\r\n"));
	       break;
            case KCMD_RFLAG_IGNOREREST:
            default:
	       /* we have nothing to do in these cases. */
                    break;
        } // Switch(iRet)
      } // if(!iExitflag)
	  return iRet;
}


