/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file kcmdSimpleInterface.h function prototypes for simple interface
  *
  */
#ifndef KCMDSIMPLEINTERFACE_H__
#define KCMDSIMPLEINTERFACE_H__

/* If we are in a c++ environment, make sure the compiler knows these are */
/* straight C functions. */
#ifdef __cplusplus
extern "C" {
#endif

/**
  * kcmd_malloc() is a wrapper for malloc() which writes a PANIC message when 
  * the underlying malloc() call fails.  It also keeps track of the current
  * amount of allocated RAM.
  */
void *kcmd_malloc( unsigned int nBytes );

/**
  * kcmd_free() is wrapper for free which watches for null pointers or crazy
  * sized blocks and will put up a PANIC message to the user if that happens.
  */
void kcmd_free(void *ptr);



/****************************************************************************
 * Here we define the bare minimum functions to make a fully functional 
 * command console.   The caller can simply provide function pointers to
 * functions that perform this to get the command processor running with 
 * a minimum of fuss.
 ****************************************************************************/

/**
  * This function waits forever for a character to arrive, and when one does,
  * it returns that character. 
  */
typedef  KCHAR (*KCMDSIMPLE_GETCH)(void);

/**
  * This function returns a zero if there are currently no characters available
  * to read, or non-zero if there is at least one character ready to be read
  * via the KCMDSIMPLE_GETCH function.
  */
typedef  int (*KCMDSIMPLE_KBHIT)(void);



/**
  * This function sends a character to the user.  It is up to the implementer
  * what they want to do if the system is not ready to send it.  It can either
  * block forever until it can send, or return and the character will be lost.
  */
typedef void (*KCMDSIMPLE_PUTCH)(KCHAR ch);


/**
  * This function sends a null-terminated string to the user.   The string
  * must be stored in RAM, since this is string of KCHAR, not a KCCHAR string.
  * Behavior if the device is not ready to send any or all of the data is up
  * to the implementer.
  */ 
typedef void (*KCMDSIMPLE_PUTSTR)(KCHAR *pStr);


/**
  * Initialize the command processor and hand over pointer to the minimal 
  * IO functions.  Call this function to get everything initialized, then 
  * after this function comes back, call the process_cmds() function in 
  * a loop to get more commands and process them.  The current state of
  * the command processor is saved in a global structure, so it is not 
  * possible to run two simultaneous command prompts with the simple
  * interface.
  */
void init_kcmd(KCCHAR *pPrompt,
               KCCHAR *pWelcomeString, 
               KCMDSIMPLE_GETCH pFuncGetch, 
			   KCMDSIMPLE_KBHIT pFuncKbHit,
			   KCMDSIMPLE_PUTCH pFuncPutch,
			   KCMDSIMPLE_PUTSTR pFuncPutStr);

/**
  * To run the command processor, call this function repeatedly in a loop.
  * When called, this function will wait for a line of input from the input
  * function.  It will then process this entire line and then return with 
  * a status.  If the status is equal to 1, then the user has indicated she
  * would like to exit the command processor, and you may end the loop.  Other
  * codes from non-fatal user errors can be acted upon, but the easiest thing
  * is to just keep calling until the return value is one.
  */
int process_kcmds(int iPromptFlag);

#ifdef __cplusplus
} /* extern "C" ... */
#endif

#endif /* KCMDSIMPLEINTERFACE_H__ */
