/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file kcmdPort_GCC_MEGA_AVR.h specific macros and definitions for
 *      the Atmel ATMEGA series chips.
 */
 #ifndef KCMDPORT_GCC_MEGA_AVR_H__
 #define KCMDPORT_GCC_MEGA_AVR_H__

/****************************************************************************
 * INCLUDES which are standard for this platform.  Include both the library
 *  and the hardware includes.
 ****************************************************************************/

/*
 * Start out with standard includes specific to this port
 */
#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>



/************************************************************************
 * CONFIG - Some parameters for configuring the command processor
 ************************************************************************/

/**
  * Maximum number of categories of commands.  Minimum is 1.
  */
#define KCMD_MAX_CATEGORIES 4


/**
  * Maximum characters in a category name 
  */
#define KCMD_MAXCATCHAR 8

/**
  * N_CMDVM_STACK - the number of elements the command stack can hold.
  */
#define N_CMDVM_STACK 8

/**
  * nLINEBUF - The maximum characters in a command line.  Keep small 
  *   in order to keep memory use down.
  */
#define nLINEBUF 80


/************************************************************************
 * TYPES - Adjust these types to be appropriate for your platform.
 ************************************************************************/
#ifndef false
#define false 0
#endif
#ifndef true
#define true 1
#endif

/**
  * KCHAR is the basic character type.  
  */
typedef char KCHAR;

/**
  * KCCHAR is the constant character suitable for storage in Program Flash 
  * memory instead of the usual RAM.
  */
typedef const char KCCHAR;

/**
 * KINTPTR_t is the type if integer which can be made into a pointer 
 * without the compiler having a hissy fit.
 */
typedef unsigned int KINTPTR_t;



/****************************************************************************
 * MACROs for storing constant text strings in program memory.   The KCCHAR
 *   type may reside in Program (Flash) memory, greatly reducing the amount
 *   of RAM the command processor needs.   Some platforms like the AVR
 *   series cannot deduce the location at runtime from the pointer itself, 
 *   so macros must be used and the functions that use these strings must
 *   know if a const char * is in RAM or in Flash.   
 ***************************************************************************/

/**
  * KCCSTR( x ) is a macro designed for systems which need string constants
  * to be stored in Flash, because of memory limitations.
  * For example:  (print_kccstr() expects the string to be in Flash)
  *    print_kccstr(pKCMD, KCCSTR("\r\n Help category not found: "));
  */
#define  KCCSTR( x )  PSTR( x )

/**
 * KCSTR(x) is a macro designed to translate text for normal string 
 * literals where KCCSTR(x) is not indicated because this string 
 * should go into normal ram.  For most platforms this macro will
 * do nothing.  But in a Windows environment with UNICODE or TCHAR
 * internationalization, it can help make kcmd using wide characters.
 */
#define KCSTR( x )  x


/**
  * A shorter alias for KCSTR( x )
  */
#define KT(x) KCSTR( x )  


/**
  *  Defines an attribute to add when defining a static array of const chars.
  *  This would be an attribute which may case the string to get saved in flash.
  *  for example 
  * For example the follow declares a string stored in Program memory.
  *   KCCHAR KCCATTR g_str_system[] =  "system"; 
  */
#define KCCATTR PROGMEM



/**
  * Defines an attribute of a global structure which should go into 
  * Flash memory.
  */
#define KCSTRUCTATTR


/**
  * when the code needs to dereference a pointer to a character which is 
  * in program space, this macro will do that. 
  */
#define KCC_ACCESS pgm_read_byte_near


/**
  * defines a string compare function where the src string is assumed to be
  * in program memory.
  */
#define KCC_strncmp( dest, src, n ) strncmp_P( dest, src, n  )

/**
  * A version of strlen which expects the src string to be in pgm memory. 
  */
#define KCC_strlen strlen_P


/**
  * A version of strcpy which expects src to be in pgm memory.
  */
#define KCC_strcpy strcpy_P




/***************************************************************************
 *  Memory allocation and Free macros
 *   The command processor uses these macros to call some wrapper functions
 *  for malloc and free.  These wrappers keep track of the total amount
 *  allocated, and there is a command 'memstat' to tell you what it is.
 *   If a malloc fails, these wrappers will also send a PANIC message to 
 *  help with debugging.  Aside from that, they could just as easily map
 *  directly to malloc() and free() library functions.
 ***************************************************************************/
#define KCMD_MALLOC(nBytes) kcmd_malloc(nBytes)
#define KCMD_FREE kcmd_free




 #endif /* KCMDPORT_GCC_MEGA_AVR_H__ */
