/*
	kcmd command processor - Copyright (C) 2009 Kevin Stokes.

	This file is part of the kcmd command processor distribution
	originally availabe from http://www.boldinventions.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the Eclipse Public License - v 1.0.

	This software has NO WARRANTY.  Use it at your own risk;   It 
	has no warranty of any kind, implied or otherwise.

	There should be a copy of the Eclipse Public License along with
	this source code.  If not, it can be found on the eclipse.org
        website:  http://www.eclipse.org

*/

/** \file user_io.h defines the USERIO container for IO functions.
 *
 *  The following function pointers define console text input and
 *  output functions.   Timeouts and IO errors are possible, but
 *  most implementations will simply always return a no-error status
 *  and will never timeout.
 *
 *  the Start, Stop, Setup, and Release functions will probably also
 *  mostly remain as do-nothing functions.  The reason for them
 *  is that this code was used in a more advanced version of the
 *  command processor which could run several sessions at once, all
 *  with their own IO methods and could spawn new sessions and free
 *  old ones.
 */
#ifndef _USER_IO_H_
#define _USER_IO_H_


#ifdef __cplusplus
extern "C" {
#endif

/**
 * USRIOFUNC_GETC
 * Get a character from user.  Returns the character that the user typed.
 * If pIErr != NULL, then a return code is stored in at that location.
 *  *pIErr=0 mean timed out before we got a char.
 *  *pIErr=1 means we got a char
 *  *pIErr<0 means error occurred
 */
typedef  KCHAR (*USRIOFUNC_GETC)(void *uio_context,  int *pIErr, int iTimeout);

/**
 * USERIOFUNC_PUTC
 *  Send one character to the user's output.   If the output buffer is full
 *  it is possible to timeout and return without having sent the character.
 *  The return value is 1 if the character was sent, 0 if a timeout occurred, 
 *  or negative if an error happened.
 */
typedef int (*USRIOFUNC_PUTC)(void *uio_context,  KCHAR inChar, int iTimeout);

/**
 * USRIOFUNC_KBHIT
 *  Check to see if there are any characters waiting in the input queue.  If
 * so, return a positive value.  If not, return a 0.  Does not remove any
 * characters from the input buffer, and does not reveal what the character is.
 */
typedef int (*USRIOFUNC_KBHIT)(void *uio_context);

/**
 * USRIOFUNC_OUTRDY
 *   Returns the output queue status.  If a positive value is returned, then
 *  there is space for at least one output character without waiting.  Zero
 *  indicates the output buffer is full and attempts to use USRIOFUNC_PUTC or 
 *  USERIOFUNC_PUTS will result in blocking if timeouts are specified.
 */
typedef int (*USRIOFUNC_OUTRDY)(void *uio_context);


/**
 * USRIOFUNC_GETS
 *   Gets a line of text from the user.   This function should collect characters
 * from the user while simultaneously echoing characters so the user can see what
 * they are typing.  When the user types a CR, the function exits and the characters
 * typed are in the buffer.   The CR character itself is not placed in the buffer.
 *   If there is no activity after iTimeout milliseconds, the function may return.
 * Backspace characters should be handled by echoing a BS then a Space then a BS,
 * to clear the last character entered.
 */
typedef int (*USRIOFUNC_GETS)(void *uio_context, KCHAR * pBuffer, int iSzBuffer, int iTimeout);


/**
 * USRIOFUNC_PUTS
 *   Puts a string of characters to the output.  Does not do any 
 * stripping of CR or newline characters; It is equivalent to calling
 * USRIOFUNC_PUTC() for each character in the null terminated input string.
 */
typedef void (*USRIOFUNC_PUTS)(void *uio_context, KCHAR * pBuffer);

/**
 * USRIOFUNC_STOP
 *  This can be called to signal the IO to stop accepting input and output.
 *  Note: Currently this is only called by the Shell before calling Release().
 *  This should always return a 1.   Many implementations of this would be to 
 * simply return 1 without stopping anything.
 */
typedef int (*USRIOFUNC_STOP)( void *uio_context);

/**
 * USRIOFUNC_START
 *  If the activity is stopped with the STOP function, this can restart it.
 * If the STOP function simply returns 1 and doesn't stop anything, then so can
 * this function.
 */
typedef int (*USRIOFUNC_START)( void *uio_context);

/**
 * USRIOFUNC_SETUP
 *  This function allocates whatever resources to make the IO operate.  It returns
 * a void pointer which must be passed for every other USRIO function.  This context
 * point might be a pointer to a malloc'd structure with serial port register info,
 * or TCP socket connection info, or whatever.  The user doesn't need to know what
 * this pointer is.
 *  The pOptionalInitParams argument is optional.  It may contain specific info that
 * the setup needs.  The KCMD shell simply passes this argument through
 */
typedef void * (*USRIOFUNC_SETUP)(void *pOptionalInitParams);

/**
 * USRIOFUNC_RELEASE
 *  This function will release any allocated resources needed by the IO.  No more
 * character IO will be possible after this function is called.   On simple implementations
 * this may also do nothing but return 1.
 */
typedef int (*USRIOFUNC_RELEASE)( void *uio_context);



/**
 * USER_IO is a container of function pointers used for
 * text console IO.
 */
typedef struct user_io_struct
{
  USRIOFUNC_GETC pFuncGetc;
  USRIOFUNC_PUTC pFuncPutc;
  USRIOFUNC_KBHIT pFuncKbhit;
  USRIOFUNC_OUTRDY pFuncOutRdy;
  USRIOFUNC_GETS pFuncGets;
  USRIOFUNC_PUTS pFuncPuts;
  USRIOFUNC_STOP pFuncStop;
  USRIOFUNC_START pFuncStart;
  USRIOFUNC_SETUP pFuncSetup;
  USRIOFUNC_RELEASE pFuncRelease;
} USER_IO;


#ifdef __cplusplus
}
#endif

#endif /* _USER_IO_H_ */
